package com.sillycycle.bagleyd.panex;

/*
 * @(#)Panex.java
 *
 * Copyright 1994 - 2020  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.Color;
import java.awt.Cursor;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import javax.swing.JFrame;

import com.sillycycle.bagleyd.panex.model.PanexFormat;
import com.sillycycle.bagleyd.panex.view.PanexGeometry;

/**
 * The <code>Panex</code> class extends the
 * <code>PanexCanvas</code> for handling the responses to the user
 * for the <code>PanexFrame</code> class.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/puzzles.html
 */

class Panex extends PanexCanvas implements MouseListener, MouseMotionListener,
		KeyListener {
	private static final long serialVersionUID = 42L;

	public Panex(JFrame frame, int mode, int tiles,
			boolean mono, boolean reverse, String names,
			Color pyramid0c, Color pyramid1c,
			Color tilec, Color bg, int delay) {
		firstPaint = true;
		this.addMouseMotionListener(this);
		this.addKeyListener(this);
		this.addMouseListener(this);
		this.frame = frame;
		if (model == null) {
			model = new PanexFormat(mode, tiles);
			geo = null;
		}
		if (geo == null) {
			geo = new PanexGeometry(model);
			aDraw = null;
		}
		geo.setMono(mono);
		geo.setReverse(reverse);
		geo.setPyramidNames(names);
		geo.setPyramidColor(0, pyramid0c);
		geo.setPyramidColor(1, pyramid1c);
		geo.setTileColor(tilec);
		geo.setBackground(bg);
		geo.setDelay(delay);
		geo.setNumberSlices((delay < PanexGeometry.MAX_SLICES)
			? delay + 1 : PanexGeometry.MAX_SLICES);
		setBackground(bg);
		setCursor(new Cursor(Cursor.HAND_CURSOR));
	}

	public boolean processKey(char key) {
		switch (key) {
		case '>':
		case '.':
			speedUpPuzzle();
			return true;
		case '<':
		case ',':
			slowDownPuzzle();
			return true;
		case '@':
			((PanexFrame) frame).toggleSound();
			return true;
		case '?':
			((PanexFrame) frame).descriptionHelp();
			return true;
		case '!':
			((PanexFrame) frame).featuresHelp();
			return true;
		case '^':
			((PanexFrame) frame).referencesHelp();
			return true;
		case 'a':
		case 'A':
			((PanexFrame) frame).aboutHelp();
			return true;
		case '*':
			((PanexFrame) frame).showBuffer();
			return true;
		default:
		}
		if (solve != null && solve.getFlag()) {
			solve.setFlag(false);
			return false;
		}
		switch (key) {
		case 'g':
			getPuzzle();
			break;
		case 'w':
			writePuzzle();
			break;
		case 'G':
			getPuzzleFile();
			break;
		case 'W':
			writePuzzleFile();
			break;
		case 'u':
		case 'U':
			undoPuzzle();
			break;
		case 'r':
		case 'R':
			redoPuzzle();
			break;
		case 'c':
		case 'C':
			clearPuzzle();
			break;
		case 's':
		case 'S':
			solvePuzzle();
			break;
		case 'm':
		case 'M':
			modePuzzle();
			break;
		case 'i':
		case 'I':
			incrementPuzzle();
			break;
		case 'd':
		case 'D':
			decrementPuzzle();
			break;
		case 'q':
		case 'Q':
		case 'x':
		case 'X':
		//case '\03': /* Control-C */
			((PanexFrame) frame).quit();
			break;
		default:
			return false;
		}
		return true;
	}

	@Override
	public void mousePressed(MouseEvent event) {
		int modifiers = event.getModifiersEx();

		if (movePaint || scanPaint || stackPaint || mouseDown)
			return;
		if (solve != null && solve.getFlag()) {
			solve.setFlag(false);
			return;
		}
		if (((modifiers & InputEvent.META_DOWN_MASK) == 0) &&
				((modifiers & InputEvent.ALT_DOWN_MASK) == 0)) {
			selectPuzzle(event.getX());
			mouseDown = true;
		} else {
			if ((event.getClickCount() & 1) == 1)
				clearWithQueryPuzzle();
			else
				clearPuzzle();
		}
	}

	@Override
	public void mouseClicked(MouseEvent event) {
		/* Method required but mousePressed does work */
	}

	@Override
	public void mouseReleased(MouseEvent event) {
		if (mouseDown) { // Do not care which mouse button
			releasePuzzle(event.getX());
			mouseDown = false;
		}
	}

	@Override
	public void mouseEntered(MouseEvent event) {
		geo.setForeground(geo.getFrameColor());
		framePaint = true;
		mouseDown = false;
		repaint();
	}

	@Override
	public void mouseExited(MouseEvent event) {
		if (mouseDown && currentStack >= 0) {
			nextStack = currentStack;
			releasePaint = true;
		}
		geo.setForeground(geo.getBorderColor());
		framePaint = true;
		mouseDown = false;
		repaint();
	}

	@Override
	public void mouseDragged(MouseEvent event) {
		motionPuzzle(event.getX());
	}

	@Override
	public void mouseMoved(MouseEvent event) {
		/* Method required but mousePressed does work */
	}

	@Override
	public void keyPressed(KeyEvent event) {
		if (event.getKeyCode() == KeyEvent.VK_ESCAPE) {
			((PanexFrame) frame).shuffleDown();
			return;
		}
		processKey(event.getKeyChar());
	}

	@Override
	public void keyReleased(KeyEvent event) {
		/* Method required but keyPressed does work */
	}

	@Override
	public void keyTyped(KeyEvent event) {
		/* Method required but keyPressed does work */
	}
}
