package com.sillycycle.bagleyd.panex;

/*
 * @(#)PanexSolve.java
 *
 * Taken from Some Results on the Panex Puzzle by Mark Manasse &
 * Danny Sleator of AT&T Bell Laboratories, and Victor K. Wei
 * of Bell Communications Research
 *
 * Thanks to Nick Baxter <nick AT baxterweb.com> for debugging level n > 4
 * and vTrick.
 *
 * Though most code by Rene Jansen <rene.j.jansen AT bigfoot.com> is now
 * removed much inspiration was gained by his efforts implementing
 * an algorithm from Quantum January/February 1996 by Vladimir
 * Dubrovsky.
 *
 * Copyright 1996 - 2019  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.Graphics;
import java.util.Random;
import javax.swing.JFrame;

import com.sillycycle.bagleyd.panex.model.PanexFormat;
import com.sillycycle.bagleyd.panex.model.PanexTile;

/**
 * The <code>Panex</code> class is the code that does the work
 * of solving the puzzle.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/puzzles.html
 */

class PanexSolve {
	private static JFrame frame;
	private static PanexCanvas panex;
	private static PanexFormat model;
	private static Random generator;
	private static final boolean DEBUG = false; //true;
	private static int counter; // debugging

	PanexSolve(JFrame frame, PanexCanvas panex) {
		PanexSolve.frame = frame;
		PanexSolve.panex = panex;
		PanexSolve.model = panex.getModel();
		PanexSolve.counter = 0;
	}

	private boolean moveATile(int fromStack, int toStack) {
		Graphics g = panex.getGraphics();

		if (g == null)
			return false;
		try {
			if (!getFlag()) {
				return false;
			}
			int fromPosition /*= panex.currentPosition*/;
			if ((fromPosition =
					panex.topOfStack(fromStack, 0)) < 0 ||
					panex.requestMove(fromStack,
					fromPosition, toStack,
					false) < 0) {
				System.err.println("move from " + fromStack +
					" to " + toStack + " cannot be made.");
				return false;
			}
			counter++;
			//System.out.println(++counter + ": " +
			//	fromStack + " " + toStack);
			//panex.currentStack = fromStack;
			//panex.nextStack = toStack;
			//panex.movePaint = true;
			//panex.paint(g);
			panex.moveTile(fromStack, fromPosition, toStack,
				PanexCanvas.DOUBLE);
			return true;
		} catch (Exception e) {
			e.printStackTrace();
			return false;
		} finally {
			g.dispose();
		}
	}

	private void solveHanoiMode() {
		boolean triangleTurn = true;
		int triangleDirection = ((model.getTiles() & 1) == 1) ? 2 : 1;
		int triangleStack = 0;
		int fromStack, toStack;
		int temp1Stack, temp2Stack, temp1Loc, temp2Loc;

		while (!panex.checkSolved()) {
			if (triangleTurn) {
				fromStack = triangleStack;
				toStack = (triangleStack + triangleDirection) %
					PanexFormat.MAX_COLUMNS;
				triangleStack = toStack;
				triangleTurn = false;
			} else {
				temp1Stack =
					(triangleStack + 1) % PanexFormat.MAX_COLUMNS;
				temp2Stack =
					(triangleStack + 2) % PanexFormat.MAX_COLUMNS;
				temp1Loc = panex.topOfStack(temp1Stack, 0);
				temp2Loc = panex.topOfStack(temp2Stack, 0);
				if (temp1Loc < 0 && temp2Loc < 0)
					return;
				else if (temp1Loc < 0) {
					fromStack = temp2Stack;
					toStack = temp1Stack;
				} else if (temp2Loc < 0) {
					fromStack = temp1Stack;
					toStack = temp2Stack;
				} else {
					PanexTile temp1Tile = panex.getModel().getColumn(temp1Stack).getTileAtPosition(temp1Loc);
					PanexTile temp2Tile = panex.getModel().getColumn(temp2Stack).getTileAtPosition(temp2Loc);
					if (temp1Tile != null && temp2Tile != null &&
							temp1Tile.getNumber() < temp2Tile.getNumber()) {
						fromStack = temp1Stack;
						toStack = temp2Stack;
					} else {
						fromStack = temp2Stack;
						toStack = temp1Stack;
					}
				}
				triangleTurn = true;
			}
			if (!moveATile(fromStack, toStack))
				return;
		}
		((PanexFrame) frame).callback(
			PanexInterface.ACTION_COMPUTED);
	}

	static void printStacks() {
		for (int position = 0; position <= model.getTiles(); position++)
			for (int stack = 0; stack < PanexFormat.MAX_COLUMNS; stack++) {
				char tileChar = '-';
				PanexTile tile = panex.getModel().getColumn(stack).getTileAtPosition(position);

				if (tile != null) {
					if (tile.getColor() == 0)
						tileChar = 'A';
					else
						tileChar = 'a';
					tileChar += (char) tile.getNumber();
				}
				System.out.print(tileChar);
				if (stack == PanexFormat.MAX_COLUMNS - 1)
					System.out.println();
				else
					System.out.print(" ");
			}
	}

	private void moveStack(int n, int a, int b) {
		int c;

		if ((a == 0 || b == 0) && (a == 1 || b == 1)) {
			c = 2;
		} else if ((a == 1 || b == 1) && (a == 2 || b == 2)) {
			c = 0;
		} else {
			c = 1;
		}
		if (n > 1)
			moveStack(n - 1, a, c);
		moveATile(a, b);
		moveATile(a, b);
		if (n > 1)
			moveStack(n - 1, c, b);
	}

	static int firstSide(boolean side) {
		return ((side) ? 2 : 0);
	}

	static int lastSide(boolean side) {
		return ((side) ? 0 : 2);
	}

	void cStack(int n, boolean side, boolean inv) {
		int first = firstSide(side);
		int last = lastSide(side);

		if (inv) {
			moveStack(n - 2, 1, first);
			moveATile(1, last);
			moveStack(n - 2, first, last);
			moveATile(1, first);
			if (n > 3)
				moveStack(n - 3, last, 1);
			moveATile(last, first);
		} else {
			moveATile(first, last);
			if (n > 3)
				moveStack(n - 3, 1, last);
			moveATile(first, 1);
			moveStack(n - 2, last, first);
			moveATile(last, 1);
			moveStack(n - 2, first, 1);
		}
	}

	void startPart(int n, boolean side, boolean inv) {
		int first = firstSide(side);
		int last = lastSide(side);
		int start = counter;

		if (inv) {
			if (n == 2) {
				moveATile(1, last);
				moveATile(1, first);
			} else if (n == 3) {
				cStack(n, side, inv);
			} else if (n > 3) {
				cStack(n, !side, inv);
				for (int i = 1; i <= (n >> 1) - 2; i++)
					cStack(n - 2 * i, side, inv);
				startPart((((n & 1) == 1) ? 3 : 2), side, inv);
			}
		} else {
			if (n == 2) {
				moveATile(first, 1);
				moveATile(last, 1);
			} else if (n == 3) {
				cStack(n, side, inv);
			} else if (n > 3) {
				startPart((((n & 1) == 1) ? 3 : 2), side, inv);
				for (int i = (n >> 1) - 2; i > 0; i--)
					cStack(n - 2 * i, side, inv);
				cStack(n, !side, inv);
			}
		}
		if (DEBUG)
			System.out.println("startPart " + (counter - start));
	}

	void finishPart(int n, boolean side, boolean inv) {
		int first = firstSide(side);
		int last = lastSide(side);
		int start = counter;

		if (inv) {
			if (n == 3) {
				moveStack(n - 2, 1, first);
				moveATile(1, last);
				moveATile(first, 1);
				moveATile(first, last);
				moveATile(1, last);
				moveATile(1, first);
				moveATile(last, first);
			} else if (n > 3) {
				moveStack(n - 2, 1, last);
				moveStack(1, 1, first);
				moveStack(n - 2, last, 1);
				moveATile(first, last);
				for (int i = 0; i <= ((n - 1) >> 1) - 2; i++)
					cStack((n - 1) - 2 * i, !side, inv);
				startPart((((n & 1) == 1) ? 2 : 3),
					!side, inv);
			}
		} else {
			if (n == 3) {
				moveATile(first, last);
				moveATile(first, 1);
				moveATile(last, 1);
				moveATile(last, first);
				moveATile(1, first);
				moveATile(last, 1);
				moveStack(n - 2, first, 1);
			} else if (n > 3) {
				startPart((((n & 1) == 1) ? 2 : 3),
					!side, inv);
				for (int i = ((n - 1) >> 1) - 2; i >= 0; i--)
					cStack((n - 1) - 2 * i, !side, inv);
				moveATile(last, first);
				moveStack(n - 2, 1, last);
				moveStack(1, first, 1);
				moveStack(n - 2, last, 1);
			}
		}
		if (DEBUG)
			System.out.println("finishPart " + (counter - start));
	}

	void churn(int n, boolean side, boolean inv) {
		int first = firstSide(side);
		int last = lastSide(side);
		int start = counter;
		/*moves = 12 * 2^(n-2)*/

		if (inv) {
			/* base free to combine */
			moveATile(last, first);

			moveStack(n - 1, 1, last);
			moveATile(first, 1);
			moveATile(first, 1);
			moveStack(n - 1, last, first);
			moveATile(1, last);
			moveATile(1, last);
			moveStack(n - 1, first, 1);

			/* separate base */
			moveATile(last, first);
		} else {
			/* base free to combine */
			moveATile(first, last);

			moveStack(n - 1, 1, first);
			moveATile(last, 1);
			moveATile(last, 1);
			moveStack(n - 1, first, last);
			moveATile(1, first);
			moveATile(1, first);
			moveStack(n - 1, last, 1);

			/* separate base */
			moveATile(first, last);
		}
		if (DEBUG)
			System.out.println("churn " + (counter - start));
	}

	void A(int n, boolean side, boolean inv) {
		int first = firstSide(side);
		int last = lastSide(side);

		if (inv) {
			if (n == 1) {
				moveATile(last, 1);
				moveATile(first, last);
				moveATile(1, first);
			} else if (n == 2) {
				startPart(n, side, !inv);
				churn(n, side, !inv);
				startPart(n, side, inv);
			} else if (n > 2) {
				finishPart(n, side, !inv);
				churn(n, side, !inv);
				startPart(n, side, inv);
			}
		} else {
			if (n == 1) {
				moveATile(first, 1);
				moveATile(last, first);
				moveATile(1, last);
			} else if (n == 2) {
				startPart(n, side, inv);
				churn(n, side, inv);
				startPart(n, side, !inv);
			} else if (n > 2) {
				startPart(n, side, inv);
				churn(n, side, inv);
				finishPart(n, side, !inv);
			}
		}
	}

	void solveAlgorithmeMode() {
		generator = new Random(System.nanoTime());
		if (model.getTiles() > 0) {
			boolean side = (generator.nextInt(2) == 1);
			boolean inv = (generator.nextInt(2) == 1);
			// Following good for debugging
			//boolean side = LEFT; // RIGHT;
			//boolean inv = NORMAL; // INVERSE;
			A(model.getTiles(), side, inv);
		}
		((PanexFrame) frame).callback(
			PanexInterface.ACTION_COMPUTED);
		counter = 0;
	}

	/*
	Recursive algorithm starts at level 5, shown by computer checking
	to be the shortest algorithm, i.e. the upper bound U(n), to at
	least to level 9.
	Lower bounds discussed by Manasse, Sleator, & Wei is given as L(n).
	T(n) = move a tower of height n
	X(n) = transfer or swap towers
	S(n) = sink a n level piece in center stack
	SQ(n) = move a single piece (for Quantum and Baxter alg.)
	DQ(n) = move a 2 n and n-1 and reorder (for Quantum and Baxter alg.)
	WB(n) = solve but one piece n in center (Baxter alg.)

	n	T(n)	X(n)	U(n)	L(n)
	1	1	3
	2	3	13
	3	9	42
	4	24	128
	5	58	343	343	320
	6	143	881	881	796
	7	345	2189	2189	1944
	8	836	5359	5359	4716
	9	2018	13023	13023	11408
	10	4875	?	31537	27564
	*/

	static final boolean LEFT = false;
	static final boolean RIGHT = true;
	static final boolean NORMAL = false;
	static final boolean INVERSE = true;

	void XSimple(int n, boolean side, boolean inv) {
		// Yeah, it works but it takes a long time
		if (n < 2)
			return;
		if (inv) {
			XSimple(n - 1, side, inv);
			TBar(n - 1, !side, !inv, false, false);
			SBar(n, !side, !inv, false);
			TBar(n - 1, side, !inv, false, false);
			SBar(n, side, !inv, false);
			SBar(n, !side, inv, false);
			TBar(n - 1, !side, inv, false, false);
			SBar(n, side, inv, false);
			TBar(n - 1, side, inv, false, false);
		} else {
			TBar(n - 1, side, inv, false, false);
			SBar(n, side, inv, false);
			TBar(n - 1, !side, inv, false, false);
			SBar(n, !side, inv, false);
			SBar(n, side, !inv, false);
			TBar(n - 1, side, !inv, false, false);
			SBar(n, !side, !inv, false);
			TBar(n - 1, !side, !inv, false, false);
			XSimple(n - 1, side, inv);
		}
	}

	// From Nick Baxter
	void XBaxter(int n, boolean side, boolean inv) {
		if (inv) {
			SQ(n, !side, !inv);
			WB(n, side, inv);
		} else {
			WB(n, side, inv);
			SQ(n, !side, !inv);
		}
	}

	// From Quantum
	void XQuantum(int n, boolean side, boolean inv) {
		int first = firstSide(side);
		int last = lastSide(side);

		if (inv) {
			SQ(n, side, !inv);
			SQ(n, !side, !inv);

			/* Original Algorithm */
			/*
			for (int i = n; i >= 3; i--)
				BQ(i, side, inv);
			moveATile(first, last);
			moveATile(1, first);
			moveATile(last, first);
			*/

			/* Rene Jansen's one move improvement */
			for (int i = n; i >= 4; i--)
				BQ(i, side, inv);
			sideSwap(side);
			moveATile(1, last);
			moveATile(first, 1);
			moveATile(first, 1);
			moveATile(last, first);
			moveATile(1, first);
			moveATile(last, first);
			moveATile(last, 1);
			moveATile(first, last);
			moveATile(1, last);
			moveATile(first, 1);
			moveATile(last, first);

			moveATile(last, first);
			if ((n & 1) == 1)
				moveATile(1, last);
			else
				moveATile(1, first);
// qTrick combine next (first, last) move to save a move if n odd
			TBar(n, !side, inv, false, ((n & 1) == 1));
		} else {
			TBar(n, !side, inv, false, ((n & 1) == 1));
// qTrick combine previous (last, first) move to save a move if n odd
			if ((n & 1) == 1)
				moveATile(last, 1);
			else
				moveATile(first, 1);
			moveATile(first, last);

			/* Original Algorithm */
			/*
			moveATile(1, last);
			moveATile(first, 1);
			moveATile(last, first);
			for (int i = 3; i <= n; i++)
				BQ(i, inv);
			*/
			/* Rene Jansen's one move improvement */
			moveATile(first, last);
			moveATile(1, first);
			moveATile(last, 1);
			moveATile(last, first);
			moveATile(1, last);
			moveATile(first, last);
			moveATile(first, 1);
			moveATile(first, last);
			moveATile(1, first);
			moveATile(1, first);
			moveATile(last, 1);
			sideSwap(side);
			for (int i = 4; i <= n; i++)
				BQ(i, side, inv);

			SQ(n, !side, !inv);
			SQ(n, side, !inv);
		}
	}

	void X(int n, boolean side, boolean inv) {
		int first = firstSide(side);
		int last = lastSide(side);

		if (n == 1) {
			if (inv) {
				T(1, !side, !inv);
				moveATile(first, last);
				T(1, side, inv);
			} else {
				T(1, side, inv);
				moveATile(last, first);
				T(1, !side, !inv);
			}
		} else if (n == 2) {
			if (inv) {
				TBar(2, !side, !inv, false, false);
				moveATile(first, last);
				moveATile(first, 1);
				moveATile(first, last);
				moveATile(1, first);
				moveATile(last, first);
				moveATile(last, 1);
				moveATile(first, last);
				moveATile(first, last);
				T(2, side, inv);
			} else {
				T(2, side, inv);
				moveATile(last, first);
				moveATile(last, first);
				moveATile(1, last);
				moveATile(first, last);
				moveATile(first, 1);
				moveATile(last, first);
				moveATile(1, first);
				moveATile(last, first);
				TBar(2, !side, !inv, false, false);
			}
		} else if (n == 3) {
			if (inv) {
				Z(side, inv);
				Y(side, inv);
				TBar(2, !side, inv, false, false);
				SBar(3, side, inv, false);
				TBar(2, side, inv, false, false);
			} else {
				TBar(2, side, inv, false, false);
				SBar(3, side, inv, false);
				TBar(2, !side, inv, false, false);
				Y(side, inv);
				Z(side, inv);
			}
		} else if (n == 4) {
			if (inv) {
				TBar(3, !side, !inv, false, false);
				SBar(3, !side, !inv, false);
				centerSwap();
				S(3, !side, inv);
				W(side, inv);
				SBar(4, side, !inv, false);
				SBar(4, !side, inv, false);
				TBar(2, !side, inv, false, false);
				TBar(2, side, !inv, false, false);
				YBar(side, inv);
				TBar(2, !side, inv, false, false);
				centerSwap();
				SBar(3, side, inv, false);
				TBar(3, side, inv, false, false);
			} else {
				TBar(3, side, inv, false, false);
				SBar(3, side, inv, false);
				centerSwap();
				TBar(2, !side, inv, false, false);
				YBar(side, inv);
				TBar(2, side, !inv, false, false);
				TBar(2, !side, inv, false, false);
				SBar(4, !side, inv, false);
				SBar(4, side, !inv, false);
				W(side, inv);
				S(3, !side, inv);
				centerSwap();
				SBar(3, !side, !inv, false);
				TBar(3, !side, !inv, false, false);
			}
		} else if (n >= 5) {
			if (inv) {
				TBar(n - 1, !side, !inv, false, false);
				V(n, side, inv);
				SBar(n, side, !inv, false);
				SBar(n, !side, inv, false);
				TBar(2, !side, inv, false, false);
				TBar(2, side, !inv, false, false);
				YBar(side, inv);
				TBar(2, !side, inv, false, false);
				TBar(2, side, !inv, false, false);
				for (int i = n - 5; i >= 1; i--) {
					SBar(n - 1 - i, side, !inv, false);
					SBar(n - 1 - i, !side, inv, false);
					TBar(n - 2 - i, !side, inv, false, false);
					TBar(n - 2 - i, side, !inv, false, false);
				}
				SBar(n - 1, side, !inv, false);
				SBar(n - 1, !side, inv, false);
				TBar(n - 2, !side, inv, false, false);
				centerSwap();
				SBar(n - 1, side, inv, false);
				TBar(n - 1, side, inv, false, false);
			} else {
				TBar(n - 1, side, inv, false, false);
				SBar(n - 1, side, inv, false);
				centerSwap();
				TBar(n - 2, !side, inv, false, false);
				SBar(n - 1, !side, inv, false);
				SBar(n - 1, side, !inv, false);
				for (int i = 1; i <= n - 5; i++) {
					TBar(n - 2 - i, side, !inv, false, false);
					TBar(n - 2 - i, !side, inv, false, false);
					SBar(n - 1 - i, !side, inv, false);
					SBar(n - 1 - i, side, !inv, false);
				}
				TBar(2, side, !inv, false, false);
				TBar(2, !side, inv, false, false);
				YBar(side, inv);
				TBar(2, side, !inv, false, false);
				TBar(2, !side, inv, false, false);
				SBar(n, !side, inv, false);
				SBar(n, side, !inv, false);
				V(n, side, inv);
				TBar(n - 1, !side, !inv, false, false);
			}
		}
	}

	/*
	n	T(n)	TBar(n)	S(n)	SBar(n)
	1	1	1	-	-
	2	3	2	2	2
	3	9	9	8	8
	4	24	23	22	22
	5	58	58	57	57
	6	143	142	141	141
	7	345	345	344	344
	8	836	835	834	834
	9	2018	2018	2017	2017
	10	4875	4874	4873	4873
	*/

	void SQ(int n, boolean side, boolean inv) {
		int first = firstSide(side);
		int last = lastSide(side);

		if (inv) {
			if (n == 1) {
				moveATile(1, first);
			} else if (n == 2) {
				moveATile(first, last);
				moveATile(1, first);
				moveATile(last, first);
			} else if (n >= 3) {
				if (((n & 1) == 0) && (generator.nextInt(2) == 1)) {
					S(n, side, inv);
					T(n - 1, side, inv);
				} else {
					SBar(n, side, inv, false);
					TBar(n - 1, side, inv, false, false);
				}
			}
		} else {
			if (n == 1) {
				moveATile(first, 1);
			} else if (n == 2) {
				moveATile(first, last);
				moveATile(first, 1);
				moveATile(last, first);
			} else if (n >= 3) {
				if (((n & 1) == 0) && (generator.nextInt(2) == 1)) {
					T(n - 1, side, inv);
					S(n, side, inv);
				} else {
					TBar(n - 1, side, inv, false, false);
					SBar(n, side, inv, false);
				}
			}
		}
	}

	void DQ(int n, boolean side, boolean inv) {
		int first = firstSide(side);
		int last = lastSide(side);

		if (n <= 1)
			return;
		if (inv) {
			if (n == 2) {
				centerSwap();
				moveATile(1, last);
				moveATile(1, first);
				moveATile(last, first);
			} else if (n >= 3) {
				centerSwap();
				if (((n & 1) == 0) && (generator.nextInt(2) == 1)) {
					S(n - 1, side, inv);
					T(n - 1, side, inv);
				} else {
					SBar(n - 1, side, inv, false);
					TBar(n - 1, side, inv, false, false);
				}
			}
		} else {
			if (n == 2) {
				moveATile(first, last);
				moveATile(first, 1);
				moveATile(last, 1);
				centerSwap();
			} else if (n >= 3) {
				if (((n & 1) == 0) && (generator.nextInt(2) == 1)) {
					T(n - 1, side, inv);
					S(n - 1, side, inv);
				} else {
					TBar(n - 1, side, inv, false, false);
					SBar(n - 1, side, inv, false);
				}
				centerSwap();
			}
		}
	}

	void WB(int n, boolean side, boolean inv) {
		int first = firstSide(side);
		int last = lastSide(side);

		if (inv) {
			if (n == 1) {
				moveATile(first, last);
				moveATile(1, first);
			} else if (n == 2) {
				sideSwap(side);
				moveATile(last, 1);
				moveATile(first, last);
				moveATile(first, last);
				moveATile(1, last);
				moveATile(1, first);
				moveATile(last, first);
			} else if (n >= 3) {
				SQ(n, side, !inv);
				SQ(n - 1, !side, !inv);
				centerSwap();
				SQ(n - 1, !side, inv);
				SQ(n - 1, side, !inv);
				SQ(n - 2, !side, !inv);
				centerSwap();
				SQ(n - 2, !side, inv);
				WB(n - 2, side, inv);
				DQ(n, side, inv);
			}
		} else {
			if (n == 1) {
				moveATile(first, 1);
				moveATile(last, first);
			} else if (n == 2) {
				moveATile(first, last);
				moveATile(first, 1);
				moveATile(last, 1);
				moveATile(last, first);
				moveATile(last, first);
				moveATile(1, last);
				sideSwap(side);
			} else if (n >= 3) {
				DQ(n, side, inv);
				WB(n - 2, side, inv);
				SQ(n - 2, !side, inv);
				centerSwap();
				SQ(n - 2, !side, !inv);
				SQ(n - 1, side, !inv);
				SQ(n - 1, !side, inv);
				centerSwap();
				SQ(n - 1, !side, !inv);
				SQ(n, side, !inv);
			}
		}
	}

	void BQ(int n, boolean side, boolean inv) {
		if (inv) {
			SQ(n - 1, side, !inv);
			centerSwap();
			SQ(n - 1, side, inv);
			SQ(n - 1, !side, !inv);
		} else {
			SQ(n - 1, !side, !inv);
			SQ(n - 1, side, inv);
			centerSwap();
			SQ(n - 1, side, !inv);
		}
	}

// Changes to TBar, SBar, V to use the "vTrick" that
// swaps B and b, rather than calling LinvTBar(3) and
// RinvTBar(3).
//
// All calls to TBar and SBar require a new parameter,
// vTrick. This will always be False, except when called
// from V, and when carefully passed down from within
// TBar and SBar. When True, the last call to LInvTBar(3)
// in LInvTBar(n-1) and the first call to RinvTBar(3)
// in RInvSBar(n-1) are suppressed. V added the replacement
// 19 moves (with 18 moves removed).

	void T(int n, boolean side, boolean inv) {
		int first = firstSide(side);
		int last = lastSide(side);

		if (inv) {
			if (n == 1) {
				moveATile(1, first);
			} else if (n == 2) {
				moveATile(1, last);
				TBar(n, side, inv, false, false);
			} else if (n == 3) {
				moveATile(1, first);
				moveATile(1, first);
				moveATile(1, last);
				moveATile(first, 1);
				moveATile(first, 1);
				moveATile(last, first);
				T(2, side, inv);
			} else if (n >= 4) {
				T(n - 2, side, inv); //Paper says invert but...
				centerSwap();
				if (((n & 1) == 0) && (generator.nextInt(2) == 1)) {
					S(n - 1, side, inv);
					T(n - 1, side, inv);
				} else {
					SBar(n - 1, side, inv, false);
					TBar(n - 1, side, inv, false, false);
				}
			}
		} else {
			if (n == 1) {
				moveATile(first, 1);
			} else if (n == 2) {
				TBar(n, side, inv, false, false);
				moveATile(last, 1);
			} else if (n == 3) {
				T(2, side, inv);
				moveATile(first, last);
				moveATile(1, first);
				moveATile(1, first);
				moveATile(last, 1);
				moveATile(first, 1);
				moveATile(first, 1);
			} else if (n >= 4) {
				if (((n & 1) == 0) && (generator.nextInt(2) == 1)) {
					T(n - 1, side, inv);
					S(n - 1, side, inv);
				} else {
					TBar(n - 1, side, inv, false, false);
					SBar(n - 1, side, inv, false);
				}
				centerSwap();
				T(n - 2, side, inv); //Paper says invert but...
			}
		}
	}

	void TBar(int n, boolean side, boolean inv, boolean vTrick, boolean qTrick) {
		int first = firstSide(side);
		int last = lastSide(side);

		if (inv) {
			if (n == 1) {
				moveATile(last, first);
			} else if (n == 2) {
				moveATile(1, first);
				moveATile(last, first);
			} else if (n == 3) {
				if (!vTrick) {
					if (!qTrick)
						moveATile(last, first);
					moveATile(1, first);
					moveATile(1, last);
					moveATile(first, 1);
					moveATile(first, 1);
					moveATile(last, first);
					T(2, side, inv);
				}
			} else if (n >= 4) {
				TBar(n - 2, side, inv, false, qTrick); //Paper says invert but...
				centerSwap();
				if (((n & 1) == 0) && (generator.nextInt(2) == 1) && !vTrick) {
					S(n - 1, side, inv);
					T(n - 1, side, inv);
				} else {
					SBar(n - 1, side, inv, false);
					TBar(n - 1, side, inv, vTrick, false);
				}
			}
		} else {
			if (n == 1) {
				moveATile(first, last);
			} else if (n == 2) {
				moveATile(first, last);
				moveATile(first, 1);
			} else if (n == 3) {
				if (!vTrick) {
					T(2, side, inv);
					moveATile(first, last);
					moveATile(1, first);
					moveATile(1, first);
					moveATile(last, 1);
					moveATile(first, 1);
					if (!qTrick)
						moveATile(first, last);
				}
			} else if (n >= 4) {
				if (((n & 1) == 0) && (generator.nextInt(2) == 1) && !vTrick) {
					T(n - 1, side, inv);
					S(n - 1, side, inv);
				} else {
					TBar(n - 1, side, inv, vTrick, false);
					SBar(n - 1, side, inv, false);
				}
				centerSwap();
				TBar(n - 2, side, inv, false, qTrick); //Paper says invert but...
			}
		}
	}

	void S(int n, boolean side, boolean inv) {
		int first = firstSide(side);
		int last = lastSide(side);

		if (inv) {
			if (n == 2) {
				SBar(n, side, inv, false);
				moveATile(last, 1);
			} else if (n == 3) {
				sideSwap(side);
				moveATile(1, last);
				moveATile(first, 1);
				moveATile(first, 1);
				moveATile(last, first);
			} else if (n >= 4) {
				if (((n & 1) == 1) && (generator.nextInt(2) == 1)) {
					T(n - 2, side, !inv);
					S(n - 1, side, !inv);
				} else {
					TBar(n - 2, side, !inv, false, false);
					SBar(n - 1, side, !inv, false);
				}
				centerSwap();
				S(n - 1, side, inv);
			}
		} else {
			if (n == 2) {
				moveATile(1, last);
				SBar(n, side, inv, false);
			} else if (n == 3) {
				moveATile(first, last);
				moveATile(1, first);
				moveATile(1, first);
				moveATile(last, 1);
				sideSwap(side);
			} else if (n >= 4) {
				S(n - 1, side, inv);
				centerSwap();
				if (((n & 1) == 1) && (generator.nextInt(2) == 1)) {
					S(n - 1, side, !inv);
					T(n - 2, side, !inv);
				} else {
					SBar(n - 1, side, !inv, false);
					TBar(n - 2, side, !inv, false, false);
				}
			}
		}
	}

	void SBar(int n, boolean side, boolean inv, boolean vTrick) {
		int first = firstSide(side);
		int last = lastSide(side);

		if (inv) {
			if (n == 2) {
				moveATile(first, last);
				moveATile(1, first);
			} else if (n == 3) {
				moveATile(first, 1);
				moveATile(first, last);
				moveATile(1, first);
				moveATile(1, first);
				moveATile(last, 1);
				moveATile(first, 1);
				moveATile(first, last);
				moveATile(1, first);
			} else if (n >= 4) {
				if (((n & 1) == 1) && (generator.nextInt(2) == 1) && !vTrick) {
					T(n - 2, side, !inv);
					S(n - 1, side, !inv);
				} else {
					TBar(n - 2, side, !inv, vTrick, false);
					SBar(n - 1, side, !inv, false);
				}
				centerSwap();
				SBar(n - 1, side, inv, false);
			}
		} else {
			if (n == 2) {
				moveATile(first, 1);
				moveATile(last, first);
			} else if (n == 3) {
				moveATile(first, 1);
				moveATile(last, first);
				moveATile(1, first);
				moveATile(1, last);
				moveATile(first, 1);
				moveATile(first, 1);
				moveATile(last, first);
				moveATile(1, first);
			} else if (n >= 4) {
				SBar(n - 1, side, inv, false);
				centerSwap();
				if (((n & 1) == 1) && (generator.nextInt(2) == 1) && !vTrick) {
					S(n - 1, side, !inv);
					T(n - 2, side, !inv);
				} else {
					SBar(n - 1, side, !inv, false);
					TBar(n - 2, side, !inv, vTrick, false);
				}
			}
		}
	}

	private void V(int n, boolean side, boolean inv) {
		if (inv) {
			SBar(n, !side, !inv, true);
			VTrick(side, inv);
			TBar(n - 1, side, !inv, true, false);
		} else {
			TBar(n - 1, side, !inv, true, false);
			VTrick(side, inv);
			SBar(n, !side, !inv, true);
		}
	}

	private void VTrick(boolean side, boolean inv) {
		int first = firstSide(side);
		int last = lastSide(side);

		if (inv) {
			moveATile(first, last);
			moveATile(1, last);
			moveATile(1, first);
			moveATile(last, 1);
			moveATile(last, 1);
			moveATile(first, last);
			W(side, inv);
		} else {
			W(side, inv);
			moveATile(last, first);
			moveATile(1, last);
			moveATile(1, last);
			moveATile(first, 1);
			moveATile(last, 1);
			moveATile(last, first);
		}
	}

	private void W(boolean side, boolean inv) {
		int first = firstSide(side);
		int last = lastSide(side);

		if (inv) {
			moveATile(first, last);
			moveATile(first, last);
			moveATile(1, first);
			moveATile(last, first);
			moveATile(last, 1);
			moveATile(first, last);
			moveATile(first, last);
			moveATile(first, last);
			moveATile(1, first);
			moveATile(1, first);
			moveATile(last, 1);
			moveATile(first, 1);
			moveATile(first, last);
		} else {
			moveATile(last, first);
			moveATile(1, first);
			moveATile(1, last);
			moveATile(first, 1);
			moveATile(first, 1);
			moveATile(last, first);
			moveATile(last, first);
			moveATile(last, first);
			moveATile(1, last);
			moveATile(first, last);
			moveATile(first, 1);
			moveATile(last, first);
			moveATile(last, first);
		}
	}

	private void Y(boolean side, boolean inv) {
		int first = firstSide(side);
		int last = lastSide(side);

		if (inv) {
			moveATile(first, last);
			moveATile(1, first);
			moveATile(last, first);
			moveATile(1, last);
			moveATile(first, 1);
			moveATile(first, 1);
			moveATile(last, first);
			moveATile(last, first);
			moveATile(last, first);
			moveATile(1, last);
			moveATile(1, last);
			moveATile(first, 1);
			moveATile(last, 1);
			moveATile(last, first);
			moveATile(1, last);
		} else {
			moveATile(last, 1);
			moveATile(first, last);
			moveATile(1, last);
			moveATile(1, first);
			moveATile(last, 1);
			moveATile(last, 1);
			moveATile(first, last);
			moveATile(first, last);
			moveATile(first, last);
			moveATile(1, first);
			moveATile(1, first);
			moveATile(last, 1);
			moveATile(first, last);
			moveATile(first, 1);
			moveATile(last, first);
		}
	}

	private void YBar(boolean side, boolean inv) {
		int first = firstSide(side);
		int last = lastSide(side);

		if (inv) {
			moveATile(first, 1);
			moveATile(last, first);
			moveATile(1, first);
			moveATile(1, last);
			moveATile(first, 1);
			moveATile(first, 1);
			moveATile(last, first);
			moveATile(last, first);
			moveATile(last, first);
			moveATile(1, last);
			moveATile(1, last);
			moveATile(first, 1);
			moveATile(last, 1);
			moveATile(last, first);
			moveATile(1, last);
		} else {
			moveATile(last, 1);
			moveATile(first, last);
			moveATile(1, last);
			moveATile(1, first);
			moveATile(last, 1);
			moveATile(last, 1);
			moveATile(first, last);
			moveATile(first, last);
			moveATile(first, last);
			moveATile(1, first);
			moveATile(1, first);
			moveATile(last, 1);
			moveATile(first, 1);
			moveATile(first, last);
			moveATile(1, first);
		}
	}

	private void Z(boolean side, boolean inv) {
		int first = firstSide(side);
		int last = lastSide(side);

		if (inv) {
			moveATile(last, first);
			moveATile(last, 1);
			moveATile(last, 1);
			moveATile(first, last);
			centerSwap();
			moveATile(first, last);
			moveATile(first, 1);
			moveATile(last, first);
			moveATile(last, first);
			moveATile(1, last);
			moveATile(first, last);
			moveATile(first, 1);
		} else {
			moveATile(1, first);
			moveATile(last, first);
			moveATile(last, 1);
			moveATile(first, last);
			moveATile(first, last);
			moveATile(1, first);
			moveATile(last, first);
			centerSwap();
			moveATile(last, first);
			moveATile(1, last);
			moveATile(1, last);
			moveATile(first, last);
		}
	}

	// Randomize to make it interesting
	private void centerSwap() {
		if (generator.nextInt(2) == 1) {
			moveATile(1, 0);
			moveATile(1, 2);
			moveATile(0, 1);
			moveATile(2, 1);
		} else {
			moveATile(1, 2);
			moveATile(1, 0);
			moveATile(2, 1);
			moveATile(0, 1);
		}
	}

	private void sideSwap(boolean side) {
		int first = firstSide(side);
		int last = lastSide(side);

		if (generator.nextInt(2) == 1) {
			moveATile(first, last);
			moveATile(first, 1);
			moveATile(last, first);
			moveATile(1, first);
		} else {
			moveATile(first, 1);
			moveATile(first, last);
			moveATile(1, first);
			moveATile(last, first);
		}
	}

	private void solvePanexMode() {
		generator = new Random(System.nanoTime());
		if (model.getTiles() > 0) {
			boolean side = (generator.nextInt(2) == 1);
			boolean inv = (generator.nextInt(2) == 1);
			// Following good for debugging
			//boolean side = LEFT; // RIGHT
			//boolean inv = NORMAL; // INVERSE
			/*if (model.getTiles() >= 3) {
				//XSimple(model.getTiles(), side, inv);
				//XBaxter(model.getTiles(), side, inv);
				XQuantum(model.getTiles(), side, inv);
			} else*/
			X(model.getTiles(), side, inv);
		}
		((PanexFrame) frame).callback(
			PanexInterface.ACTION_COMPUTED);
		counter = 0;
	}

	private void solveTiles() {
		if (model.getMode() == PanexFormat.Modes.PANEX.ordinal()) {
			solvePanexMode();
		} else if (model.getMode() == PanexFormat.Modes.ALGORITHME.ordinal()) {
			solveAlgorithmeMode();
		} else /*if (panex.mode == PanexFormat.Modes.HANOI.ordinal())*/ {
			solveHanoiMode();
		}
	}

	private class SeparateSubTask extends Thread {
		private boolean runFlag = false;
		SeparateSubTask() {
		}
		public void init() {
			start();
		}
		public void setFlag(boolean flag) {
			runFlag = flag;
		}
		public boolean getFlag() {
			return runFlag;
		}

		@Override
		public void run() {
			while (true) {
				try {
					sleep(100);
				} catch (InterruptedException e) {
					System.err.println("Interrupted");
				}
				if (runFlag) {
					solveTiles();
					runFlag = false;
				}
			}
		}
	}

	private SeparateSubTask sp = null;
	public void setFlag(boolean flag) {
		if (sp != null)
			sp.setFlag(flag);
	}
	public boolean getFlag() {
		if (sp == null)
			return false;
		return sp.getFlag();
	}
	public void init() {
		if (sp == null) {
			sp = new SeparateSubTask();
			sp.init();
		}
	}
}
